COPY connectsTo     FROM 'PATHVAR/connectsTo.csv';
COPY monitoredBy    FROM 'PATHVAR/monitoredBy.csv';
COPY Region         FROM 'PATHVAR/Region.csv';
COPY requires       FROM 'PATHVAR/requires.csv';
COPY Route          FROM 'PATHVAR/Route.csv';
COPY Segment        FROM 'PATHVAR/Segment.csv';
COPY Semaphore      FROM 'PATHVAR/Semaphore.csv';
COPY Sensor         FROM 'PATHVAR/Sensor.csv';
COPY Switch         FROM 'PATHVAR/Switch.csv';
COPY SwitchPosition FROM 'PATHVAR/SwitchPosition.csv';
COPY TrackElement   FROM 'PATHVAR/TrackElement.csv';
