INSERT INTO aka_name SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_aka_name.parquet';
INSERT INTO aka_title SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_aka_title.parquet';
INSERT INTO cast_info SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_cast_info.parquet';
INSERT INTO char_name SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_char_name.parquet';
INSERT INTO comp_cast_type SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_comp_cast_type.parquet';
INSERT INTO company_name SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_company_name.parquet';
INSERT INTO company_type SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_company_type.parquet';
INSERT INTO complete_cast SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_complete_cast.parquet';
INSERT INTO info_type SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_info_type.parquet';
INSERT INTO keyword SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_keyword.parquet';
INSERT INTO kind_type SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_kind_type.parquet';
INSERT INTO link_type SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_link_type.parquet';
INSERT INTO movie_companies SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_movie_companies.parquet';
INSERT INTO movie_info SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_movie_info.parquet';
INSERT INTO movie_info_idx SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_movie_info_idx.parquet';
INSERT INTO movie_keyword SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_movie_keyword.parquet';
INSERT INTO movie_link SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_movie_link.parquet';
INSERT INTO name SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_name.parquet';
INSERT INTO person_info SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_person_info.parquet';
INSERT INTO role_type SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_role_type.parquet';
INSERT INTO title SELECT * FROM 'https://github.com/duckdb/duckdb-data/releases/download/v1.0/job_title.parquet';
