/*:
 [Home](1.%20Welcome)
 # Welcome to DuckDB
 
 ## Introduction 
 DuckDB is a high-performance analytical database system. It is designed to be
 fast, reliable and easy to use. DuckDB provides a rich SQL dialect, with
 support far beyond basic SQL. DuckDB supports arbitrary and nested correlated
 subqueries, window functions, collations, complex types (arrays, structs), and
 more. For more information on the goals of DuckDB, please refer to the [Why
 DuckDB page](https://duckdb.org/why_duckdb) on our website.
 
 ## How to use these playgrounds
 Each playground is a live code example that you can run and edit to familiarise
 yourself with the DuckDB Swift API. Playgrounds should run automatically, but
 if your environment is set-up to run manually, press the play button in the
 debug window to get started.
 
 > For playgrounds in this series to work, it's important to access them via the
 > `DuckDB.xcworkspace` file otherwise Xcode Playgrounds will be unable to find
 > the DuckDB Swift module.
 
 ## Table of Contents
 1. [Welcome to DuckDB](1.%20Welcome) (This page)
 2. [Getting Started](2.%20Getting%20Started)
 3. [Visualizing Data](3.%20Visualizing%20Data)
 
 */
