/* THIS FILE WAS AUTOMATICALLY GENERATED BY inline_grammar.py */
#include "keyword_helper.hpp"

namespace duckdb {
void PEGKeywordHelper::InitializeKeywordMaps() { // Renamed for clarity
	if (initialized) {
		return;
	};
	initialized = true;

	// Populating reserved_keyword_map
	reserved_keyword_map.insert("all");
	reserved_keyword_map.insert("analyse");
	reserved_keyword_map.insert("analyze");
	reserved_keyword_map.insert("and");
	reserved_keyword_map.insert("any");
	reserved_keyword_map.insert("array");
	reserved_keyword_map.insert("as");
	reserved_keyword_map.insert("asc");
	reserved_keyword_map.insert("asymmetric");
	reserved_keyword_map.insert("both");
	reserved_keyword_map.insert("case");
	reserved_keyword_map.insert("cast");
	reserved_keyword_map.insert("check");
	reserved_keyword_map.insert("collate");
	reserved_keyword_map.insert("column");
	reserved_keyword_map.insert("constraint");
	reserved_keyword_map.insert("create");
	reserved_keyword_map.insert("default");
	reserved_keyword_map.insert("deferrable");
	reserved_keyword_map.insert("desc");
	reserved_keyword_map.insert("describe");
	reserved_keyword_map.insert("distinct");
	reserved_keyword_map.insert("do");
	reserved_keyword_map.insert("else");
	reserved_keyword_map.insert("end");
	reserved_keyword_map.insert("except");
	reserved_keyword_map.insert("false");
	reserved_keyword_map.insert("fetch");
	reserved_keyword_map.insert("for");
	reserved_keyword_map.insert("foreign");
	reserved_keyword_map.insert("from");
	reserved_keyword_map.insert("group");
	reserved_keyword_map.insert("having");
	reserved_keyword_map.insert("in");
	reserved_keyword_map.insert("initially");
	reserved_keyword_map.insert("intersect");
	reserved_keyword_map.insert("into");
	reserved_keyword_map.insert("lambda");
	reserved_keyword_map.insert("lateral");
	reserved_keyword_map.insert("leading");
	reserved_keyword_map.insert("limit");
	reserved_keyword_map.insert("not");
	reserved_keyword_map.insert("null");
	reserved_keyword_map.insert("offset");
	reserved_keyword_map.insert("on");
	reserved_keyword_map.insert("only");
	reserved_keyword_map.insert("or");
	reserved_keyword_map.insert("order");
	reserved_keyword_map.insert("pivot");
	reserved_keyword_map.insert("pivot_longer");
	reserved_keyword_map.insert("pivot_wider");
	reserved_keyword_map.insert("placing");
	reserved_keyword_map.insert("primary");
	reserved_keyword_map.insert("qualify");
	reserved_keyword_map.insert("references");
	reserved_keyword_map.insert("returning");
	reserved_keyword_map.insert("select");
	reserved_keyword_map.insert("show");
	reserved_keyword_map.insert("some");
	reserved_keyword_map.insert("summarize");
	reserved_keyword_map.insert("symmetric");
	reserved_keyword_map.insert("table");
	reserved_keyword_map.insert("then");
	reserved_keyword_map.insert("to");
	reserved_keyword_map.insert("trailing");
	reserved_keyword_map.insert("true");
	reserved_keyword_map.insert("union");
	reserved_keyword_map.insert("unique");
	reserved_keyword_map.insert("unpivot");
	reserved_keyword_map.insert("using");
	reserved_keyword_map.insert("variadic");
	reserved_keyword_map.insert("when");
	reserved_keyword_map.insert("where");
	reserved_keyword_map.insert("window");
	reserved_keyword_map.insert("with");

	// Populating unreserved_keyword_map
	unreserved_keyword_map.insert("abort");
	unreserved_keyword_map.insert("absolute");
	unreserved_keyword_map.insert("access");
	unreserved_keyword_map.insert("action");
	unreserved_keyword_map.insert("add");
	unreserved_keyword_map.insert("admin");
	unreserved_keyword_map.insert("after");
	unreserved_keyword_map.insert("aggregate");
	unreserved_keyword_map.insert("also");
	unreserved_keyword_map.insert("alter");
	unreserved_keyword_map.insert("always");
	unreserved_keyword_map.insert("assertion");
	unreserved_keyword_map.insert("assignment");
	unreserved_keyword_map.insert("attach");
	unreserved_keyword_map.insert("attribute");
	unreserved_keyword_map.insert("backward");
	unreserved_keyword_map.insert("before");
	unreserved_keyword_map.insert("begin");
	unreserved_keyword_map.insert("cache");
	unreserved_keyword_map.insert("call");
	unreserved_keyword_map.insert("called");
	unreserved_keyword_map.insert("cascade");
	unreserved_keyword_map.insert("cascaded");
	unreserved_keyword_map.insert("catalog");
	unreserved_keyword_map.insert("centuries");
	unreserved_keyword_map.insert("century");
	unreserved_keyword_map.insert("chain");
	unreserved_keyword_map.insert("characteristics");
	unreserved_keyword_map.insert("checkpoint");
	unreserved_keyword_map.insert("class");
	unreserved_keyword_map.insert("close");
	unreserved_keyword_map.insert("cluster");
	unreserved_keyword_map.insert("comment");
	unreserved_keyword_map.insert("comments");
	unreserved_keyword_map.insert("commit");
	unreserved_keyword_map.insert("committed");
	unreserved_keyword_map.insert("compression");
	unreserved_keyword_map.insert("configuration");
	unreserved_keyword_map.insert("conflict");
	unreserved_keyword_map.insert("connection");
	unreserved_keyword_map.insert("constraints");
	unreserved_keyword_map.insert("content");
	unreserved_keyword_map.insert("continue");
	unreserved_keyword_map.insert("conversion");
	unreserved_keyword_map.insert("copy");
	unreserved_keyword_map.insert("cost");
	unreserved_keyword_map.insert("csv");
	unreserved_keyword_map.insert("cube");
	unreserved_keyword_map.insert("current");
	unreserved_keyword_map.insert("cursor");
	unreserved_keyword_map.insert("cycle");
	unreserved_keyword_map.insert("data");
	unreserved_keyword_map.insert("database");
	unreserved_keyword_map.insert("day");
	unreserved_keyword_map.insert("days");
	unreserved_keyword_map.insert("deallocate");
	unreserved_keyword_map.insert("decade");
	unreserved_keyword_map.insert("decades");
	unreserved_keyword_map.insert("declare");
	unreserved_keyword_map.insert("defaults");
	unreserved_keyword_map.insert("deferred");
	unreserved_keyword_map.insert("definer");
	unreserved_keyword_map.insert("delete");
	unreserved_keyword_map.insert("delimiter");
	unreserved_keyword_map.insert("delimiters");
	unreserved_keyword_map.insert("depends");
	unreserved_keyword_map.insert("detach");
	unreserved_keyword_map.insert("dictionary");
	unreserved_keyword_map.insert("disable");
	unreserved_keyword_map.insert("discard");
	unreserved_keyword_map.insert("document");
	unreserved_keyword_map.insert("domain");
	unreserved_keyword_map.insert("double");
	unreserved_keyword_map.insert("drop");
	unreserved_keyword_map.insert("each");
	unreserved_keyword_map.insert("enable");
	unreserved_keyword_map.insert("encoding");
	unreserved_keyword_map.insert("encrypted");
	unreserved_keyword_map.insert("enum");
	unreserved_keyword_map.insert("error");
	unreserved_keyword_map.insert("escape");
	unreserved_keyword_map.insert("event");
	unreserved_keyword_map.insert("exclude");
	unreserved_keyword_map.insert("excluding");
	unreserved_keyword_map.insert("exclusive");
	unreserved_keyword_map.insert("execute");
	unreserved_keyword_map.insert("explain");
	unreserved_keyword_map.insert("export");
	unreserved_keyword_map.insert("export_state");
	unreserved_keyword_map.insert("extension");
	unreserved_keyword_map.insert("extensions");
	unreserved_keyword_map.insert("external");
	unreserved_keyword_map.insert("family");
	unreserved_keyword_map.insert("filter");
	unreserved_keyword_map.insert("first");
	unreserved_keyword_map.insert("following");
	unreserved_keyword_map.insert("force");
	unreserved_keyword_map.insert("forward");
	unreserved_keyword_map.insert("function");
	unreserved_keyword_map.insert("functions");
	unreserved_keyword_map.insert("global");
	unreserved_keyword_map.insert("grant");
	unreserved_keyword_map.insert("granted");
	unreserved_keyword_map.insert("groups");
	unreserved_keyword_map.insert("handler");
	unreserved_keyword_map.insert("header");
	unreserved_keyword_map.insert("hold");
	unreserved_keyword_map.insert("hour");
	unreserved_keyword_map.insert("hours");
	unreserved_keyword_map.insert("identity");
	unreserved_keyword_map.insert("if");
	unreserved_keyword_map.insert("ignore");
	unreserved_keyword_map.insert("immediate");
	unreserved_keyword_map.insert("immutable");
	unreserved_keyword_map.insert("implicit");
	unreserved_keyword_map.insert("import");
	unreserved_keyword_map.insert("include");
	unreserved_keyword_map.insert("including");
	unreserved_keyword_map.insert("increment");
	unreserved_keyword_map.insert("index");
	unreserved_keyword_map.insert("indexes");
	unreserved_keyword_map.insert("inherit");
	unreserved_keyword_map.insert("inherits");
	unreserved_keyword_map.insert("inline");
	unreserved_keyword_map.insert("input");
	unreserved_keyword_map.insert("insensitive");
	unreserved_keyword_map.insert("insert");
	unreserved_keyword_map.insert("install");
	unreserved_keyword_map.insert("instead");
	unreserved_keyword_map.insert("invoker");
	unreserved_keyword_map.insert("isolation");
	unreserved_keyword_map.insert("json");
	unreserved_keyword_map.insert("key");
	unreserved_keyword_map.insert("label");
	unreserved_keyword_map.insert("language");
	unreserved_keyword_map.insert("large");
	unreserved_keyword_map.insert("last");
	unreserved_keyword_map.insert("leakproof");
	unreserved_keyword_map.insert("level");
	unreserved_keyword_map.insert("listen");
	unreserved_keyword_map.insert("load");
	unreserved_keyword_map.insert("local");
	unreserved_keyword_map.insert("location");
	unreserved_keyword_map.insert("lock");
	unreserved_keyword_map.insert("locked");
	unreserved_keyword_map.insert("logged");
	unreserved_keyword_map.insert("macro");
	unreserved_keyword_map.insert("mapping");
	unreserved_keyword_map.insert("match");
	unreserved_keyword_map.insert("matched");
	unreserved_keyword_map.insert("materialized");
	unreserved_keyword_map.insert("maxvalue");
	unreserved_keyword_map.insert("merge");
	unreserved_keyword_map.insert("method");
	unreserved_keyword_map.insert("microsecond");
	unreserved_keyword_map.insert("microseconds");
	unreserved_keyword_map.insert("millennia");
	unreserved_keyword_map.insert("millennium");
	unreserved_keyword_map.insert("millisecond");
	unreserved_keyword_map.insert("milliseconds");
	unreserved_keyword_map.insert("minute");
	unreserved_keyword_map.insert("minutes");
	unreserved_keyword_map.insert("minvalue");
	unreserved_keyword_map.insert("mode");
	unreserved_keyword_map.insert("month");
	unreserved_keyword_map.insert("months");
	unreserved_keyword_map.insert("move");
	unreserved_keyword_map.insert("name");
	unreserved_keyword_map.insert("names");
	unreserved_keyword_map.insert("new");
	unreserved_keyword_map.insert("next");
	unreserved_keyword_map.insert("no");
	unreserved_keyword_map.insert("nothing");
	unreserved_keyword_map.insert("notify");
	unreserved_keyword_map.insert("nowait");
	unreserved_keyword_map.insert("nulls");
	unreserved_keyword_map.insert("object");
	unreserved_keyword_map.insert("of");
	unreserved_keyword_map.insert("off");
	unreserved_keyword_map.insert("oids");
	unreserved_keyword_map.insert("old");
	unreserved_keyword_map.insert("operator");
	unreserved_keyword_map.insert("option");
	unreserved_keyword_map.insert("options");
	unreserved_keyword_map.insert("ordinality");
	unreserved_keyword_map.insert("others");
	unreserved_keyword_map.insert("over");
	unreserved_keyword_map.insert("overriding");
	unreserved_keyword_map.insert("owned");
	unreserved_keyword_map.insert("owner");
	unreserved_keyword_map.insert("parallel");
	unreserved_keyword_map.insert("parser");
	unreserved_keyword_map.insert("partial");
	unreserved_keyword_map.insert("partition");
	unreserved_keyword_map.insert("partitioned");
	unreserved_keyword_map.insert("passing");
	unreserved_keyword_map.insert("password");
	unreserved_keyword_map.insert("percent");
	unreserved_keyword_map.insert("persistent");
	unreserved_keyword_map.insert("plans");
	unreserved_keyword_map.insert("policy");
	unreserved_keyword_map.insert("pragma");
	unreserved_keyword_map.insert("preceding");
	unreserved_keyword_map.insert("prepare");
	unreserved_keyword_map.insert("prepared");
	unreserved_keyword_map.insert("preserve");
	unreserved_keyword_map.insert("prior");
	unreserved_keyword_map.insert("privileges");
	unreserved_keyword_map.insert("procedural");
	unreserved_keyword_map.insert("procedure");
	unreserved_keyword_map.insert("program");
	unreserved_keyword_map.insert("publication");
	unreserved_keyword_map.insert("quarter");
	unreserved_keyword_map.insert("quarters");
	unreserved_keyword_map.insert("quote");
	unreserved_keyword_map.insert("range");
	unreserved_keyword_map.insert("read");
	unreserved_keyword_map.insert("reassign");
	unreserved_keyword_map.insert("recheck");
	unreserved_keyword_map.insert("recursive");
	unreserved_keyword_map.insert("ref");
	unreserved_keyword_map.insert("referencing");
	unreserved_keyword_map.insert("refresh");
	unreserved_keyword_map.insert("reindex");
	unreserved_keyword_map.insert("relative");
	unreserved_keyword_map.insert("release");
	unreserved_keyword_map.insert("rename");
	unreserved_keyword_map.insert("repeatable");
	unreserved_keyword_map.insert("replace");
	unreserved_keyword_map.insert("replica");
	unreserved_keyword_map.insert("reset");
	unreserved_keyword_map.insert("respect");
	unreserved_keyword_map.insert("restart");
	unreserved_keyword_map.insert("restrict");
	unreserved_keyword_map.insert("returns");
	unreserved_keyword_map.insert("revoke");
	unreserved_keyword_map.insert("role");
	unreserved_keyword_map.insert("rollback");
	unreserved_keyword_map.insert("rollup");
	unreserved_keyword_map.insert("rows");
	unreserved_keyword_map.insert("rule");
	unreserved_keyword_map.insert("sample");
	unreserved_keyword_map.insert("savepoint");
	unreserved_keyword_map.insert("schema");
	unreserved_keyword_map.insert("schemas");
	unreserved_keyword_map.insert("scope");
	unreserved_keyword_map.insert("scroll");
	unreserved_keyword_map.insert("search");
	unreserved_keyword_map.insert("second");
	unreserved_keyword_map.insert("seconds");
	unreserved_keyword_map.insert("secret");
	unreserved_keyword_map.insert("security");
	unreserved_keyword_map.insert("sequence");
	unreserved_keyword_map.insert("sequences");
	unreserved_keyword_map.insert("serializable");
	unreserved_keyword_map.insert("server");
	unreserved_keyword_map.insert("session");
	unreserved_keyword_map.insert("set");
	unreserved_keyword_map.insert("sets");
	unreserved_keyword_map.insert("share");
	unreserved_keyword_map.insert("simple");
	unreserved_keyword_map.insert("skip");
	unreserved_keyword_map.insert("snapshot");
	unreserved_keyword_map.insert("sorted");
	unreserved_keyword_map.insert("source");
	unreserved_keyword_map.insert("sql");
	unreserved_keyword_map.insert("stable");
	unreserved_keyword_map.insert("standalone");
	unreserved_keyword_map.insert("start");
	unreserved_keyword_map.insert("statement");
	unreserved_keyword_map.insert("statistics");
	unreserved_keyword_map.insert("stdin");
	unreserved_keyword_map.insert("stdout");
	unreserved_keyword_map.insert("storage");
	unreserved_keyword_map.insert("stored");
	unreserved_keyword_map.insert("strict");
	unreserved_keyword_map.insert("strip");
	unreserved_keyword_map.insert("subscription");
	unreserved_keyword_map.insert("sysid");
	unreserved_keyword_map.insert("system");
	unreserved_keyword_map.insert("tables");
	unreserved_keyword_map.insert("tablespace");
	unreserved_keyword_map.insert("target");
	unreserved_keyword_map.insert("temp");
	unreserved_keyword_map.insert("template");
	unreserved_keyword_map.insert("temporary");
	unreserved_keyword_map.insert("text");
	unreserved_keyword_map.insert("ties");
	unreserved_keyword_map.insert("transaction");
	unreserved_keyword_map.insert("transform");
	unreserved_keyword_map.insert("trigger");
	unreserved_keyword_map.insert("truncate");
	unreserved_keyword_map.insert("trusted");
	unreserved_keyword_map.insert("type");
	unreserved_keyword_map.insert("types");
	unreserved_keyword_map.insert("unbounded");
	unreserved_keyword_map.insert("uncommitted");
	unreserved_keyword_map.insert("unencrypted");
	unreserved_keyword_map.insert("unknown");
	unreserved_keyword_map.insert("unlisten");
	unreserved_keyword_map.insert("unlogged");
	unreserved_keyword_map.insert("until");
	unreserved_keyword_map.insert("update");
	unreserved_keyword_map.insert("use");
	unreserved_keyword_map.insert("user");
	unreserved_keyword_map.insert("vacuum");
	unreserved_keyword_map.insert("valid");
	unreserved_keyword_map.insert("validate");
	unreserved_keyword_map.insert("validator");
	unreserved_keyword_map.insert("value");
	unreserved_keyword_map.insert("variable");
	unreserved_keyword_map.insert("varying");
	unreserved_keyword_map.insert("version");
	unreserved_keyword_map.insert("view");
	unreserved_keyword_map.insert("views");
	unreserved_keyword_map.insert("virtual");
	unreserved_keyword_map.insert("volatile");
	unreserved_keyword_map.insert("week");
	unreserved_keyword_map.insert("weeks");
	unreserved_keyword_map.insert("whitespace");
	unreserved_keyword_map.insert("within");
	unreserved_keyword_map.insert("without");
	unreserved_keyword_map.insert("work");
	unreserved_keyword_map.insert("wrapper");
	unreserved_keyword_map.insert("write");
	unreserved_keyword_map.insert("xml");
	unreserved_keyword_map.insert("year");
	unreserved_keyword_map.insert("years");
	unreserved_keyword_map.insert("yes");
	unreserved_keyword_map.insert("zone");

	// Populating colname_keyword_map
	colname_keyword_map.insert("between");
	colname_keyword_map.insert("bigint");
	colname_keyword_map.insert("bit");
	colname_keyword_map.insert("boolean");
	colname_keyword_map.insert("char");
	colname_keyword_map.insert("character");
	colname_keyword_map.insert("coalesce");
	colname_keyword_map.insert("columns");
	colname_keyword_map.insert("dec");
	colname_keyword_map.insert("decimal");
	colname_keyword_map.insert("exists");
	colname_keyword_map.insert("extract");
	colname_keyword_map.insert("float");
	colname_keyword_map.insert("generated");
	colname_keyword_map.insert("grouping");
	colname_keyword_map.insert("grouping_id");
	colname_keyword_map.insert("inout");
	colname_keyword_map.insert("int");
	colname_keyword_map.insert("integer");
	colname_keyword_map.insert("interval");
	colname_keyword_map.insert("map");
	colname_keyword_map.insert("national");
	colname_keyword_map.insert("nchar");
	colname_keyword_map.insert("none");
	colname_keyword_map.insert("nullif");
	colname_keyword_map.insert("numeric");
	colname_keyword_map.insert("out");
	colname_keyword_map.insert("overlay");
	colname_keyword_map.insert("position");
	colname_keyword_map.insert("precision");
	colname_keyword_map.insert("real");
	colname_keyword_map.insert("row");
	colname_keyword_map.insert("setof");
	colname_keyword_map.insert("smallint");
	colname_keyword_map.insert("struct");
	colname_keyword_map.insert("substring");
	colname_keyword_map.insert("time");
	colname_keyword_map.insert("timestamp");
	colname_keyword_map.insert("treat");
	colname_keyword_map.insert("trim");
	colname_keyword_map.insert("try_cast");
	colname_keyword_map.insert("values");
	colname_keyword_map.insert("varchar");
	colname_keyword_map.insert("xmlattributes");
	colname_keyword_map.insert("xmlconcat");
	colname_keyword_map.insert("xmlelement");
	colname_keyword_map.insert("xmlexists");
	colname_keyword_map.insert("xmlforest");
	colname_keyword_map.insert("xmlnamespaces");
	colname_keyword_map.insert("xmlparse");
	colname_keyword_map.insert("xmlpi");
	colname_keyword_map.insert("xmlroot");
	colname_keyword_map.insert("xmlserialize");
	colname_keyword_map.insert("xmltable");

	// Populating typefunc_keyword_map
	typefunc_keyword_map.insert("anti");
	typefunc_keyword_map.insert("asof");
	typefunc_keyword_map.insert("at");
	typefunc_keyword_map.insert("authorization");
	typefunc_keyword_map.insert("binary");
	typefunc_keyword_map.insert("by");
	typefunc_keyword_map.insert("collation");
	typefunc_keyword_map.insert("columns");
	typefunc_keyword_map.insert("concurrently");
	typefunc_keyword_map.insert("cross");
	typefunc_keyword_map.insert("freeze");
	typefunc_keyword_map.insert("full");
	typefunc_keyword_map.insert("generated");
	typefunc_keyword_map.insert("glob");
	typefunc_keyword_map.insert("ilike");
	typefunc_keyword_map.insert("inner");
	typefunc_keyword_map.insert("is");
	typefunc_keyword_map.insert("isnull");
	typefunc_keyword_map.insert("join");
	typefunc_keyword_map.insert("left");
	typefunc_keyword_map.insert("like");
	typefunc_keyword_map.insert("map");
	typefunc_keyword_map.insert("natural");
	typefunc_keyword_map.insert("notnull");
	typefunc_keyword_map.insert("outer");
	typefunc_keyword_map.insert("overlaps");
	typefunc_keyword_map.insert("positional");
	typefunc_keyword_map.insert("right");
	typefunc_keyword_map.insert("semi");
	typefunc_keyword_map.insert("similar");
	typefunc_keyword_map.insert("struct");
	typefunc_keyword_map.insert("tablesample");
	typefunc_keyword_map.insert("try_cast");
	typefunc_keyword_map.insert("unpack");
	typefunc_keyword_map.insert("verbose");
}
} // namespace duckdb
