#include <windows.h>

#define Q(x) #x
#define QUOTE(x) Q(x)

VS_VERSION_INFO VERSIONINFO
	FILEVERSION DUCKDB_MAJOR_VERSION,DUCKDB_MINOR_VERSION,DUCKDB_PATCH_VERSION,DUCKDB_DEV_ITERATION
	PRODUCTVERSION DUCKDB_MAJOR_VERSION,DUCKDB_MINOR_VERSION,DUCKDB_PATCH_VERSION,DUCKDB_DEV_ITERATION
#ifdef DEBUG
	FILEFLAGSMASK VS_FF_DEBUG | VS_FF_PRERELEASE
#else
	FILEFLAGSMASK 0
#endif
	FILEOS VOS_NT_WINDOWS32
	FILETYPE VFT_APP
BEGIN
	BLOCK "StringFileInfo"
	BEGIN
		BLOCK "040904b0"
		BEGIN
			VALUE "Comments",         "DuckDB shell"
			VALUE "CompanyName",      "DuckDB Labs"

			VALUE "FileDescription",  "DuckDB shell"
			VALUE "FileVersion",      QUOTE(DUCKDB_VERSION)
			VALUE "InternalName",     "DuckDB shell"
			VALUE "LegalCopyright",   "Copyright 2018-" QUOTE(DUCKDB_COPYRIGHT_YEAR) " Stichting DuckDB Foundation"

			VALUE "OriginalFilename", "duckdb.exe"
			VALUE "ProductName",      "DuckDB"
			VALUE "ProductVersion",   QUOTE(DUCKDB_VERSION)
		END
	END
	BLOCK "VarFileInfo"
	BEGIN
		VALUE "Translation", 0x409, 1252
	END
END

MAINICON ICON "../../../logo/DuckDB.ico"
