CALL dbgen(sf=100, children=10, step=0);
CALL dbgen(sf=100, children=10, step=1);
CALL dbgen(sf=100, children=10, step=2);
CALL dbgen(sf=100, children=10, step=3);
CALL dbgen(sf=100, children=10, step=4);
CALL dbgen(sf=100, children=10, step=5);
CALL dbgen(sf=100, children=10, step=6);
CALL dbgen(sf=100, children=10, step=7);
CALL dbgen(sf=100, children=10, step=8);
CALL dbgen(sf=100, children=10, step=9);